"""
This subpackage contains code to parse search queries received from the
frontend into trees that can be used by the database backend.
"""

from __future__ import unicode_literals
from shlex import split

from dateutil.parser import parse as parse_date

from .nodes import (String, Regex, Text, Language, Author, Date, Symbol,
                    BinaryOp, UnaryOp)
from .tree import Tree
from ..languages import LANGS

__all__ = ["QueryParseException", "parse_query"]

class QueryParseException(Exception):
    """Raised by parse_query() when a query is invalid."""
    pass


class _QueryParser(object):
    """Wrapper class with methods to parse queries. Used as a singleton."""

    def __init__(self):
        self._prefixes = {
            self._parse_language: ["l", "lang", "language"],
            self._parse_author: ["a", "author"],
            self._parse_modified: ["m", "mod", "modified", "modify"],
            self._parse_created: ["cr", "create", "created"],
            self._parse_symbol: ["s", "sym", "symb", "symbol"],
            self._parse_function: ["f", "fn", "func", "function"],
            self._parse_class: ["cl", "class", "clss"],
            self._parse_variable: ["v", "var", "variable"]
        }

    def _parse_language(self, term):
        """Parse part of a query into a language node and return it."""
        ## TODO: look up language ID
        return Language(0)

    def _parse_author(self, term):
        """Parse part of a query into an author node and return it."""
        return Author(self._parse_literal(term))

    def _parse_date(self, term, type_):
        """Parse part of a query into a date node and return it."""
        if term.startswith(("before:", "b:")):
            relation = Date.BEFORE
            dtstr = term.split(":", 1)[1]
        elif term.startswith(("after:", "a:")):
            relation = Date.AFTER
            dtstr = term.split(":", 1)[1]
        else:
            raise QueryParseException('Bad relation for date: "%s"' % term)
        try:
            dt = parse_date(dtstr)
        except (TypeError, ValueError):
            raise QueryParseException('Bad datetime for date: "%s"' % dtstr)
        return Date(type_, relation, dt)

    def _parse_modified(self, term):
        """Parse part of a query into a date modified node and return it."""
        return self._parse_date(term, Date.MODIFY)

    def _parse_created(self, term):
        """Parse part of a query into a date created node and return it."""
        return self._parse_date(term, Date.CREATE)

    def _parse_symbol(self, term):
        """Parse part of a query into a symbol node and return it."""
        pass

    def _parse_function(self, term):
        """Parse part of a query into a function node and return it."""
        pass

    def _parse_class(self, term):
        """Parse part of a query into a class node and return it."""
        pass

    def _parse_variable(self, term):
        """Parse part of a query into a variable node and return it."""
        pass

    def _parse_literal(self, literal):
        """Parse part of a search query into a string or regular expression."""
        if literal.startswith(("r:", "re:", "regex:", "regexp:")):
            return Regex(literal.split(":", 1)[1])
        return String(literal)

    def _parse_term(self, term):
        """Parse a query term into a tree node and return it."""
        if ":" in term and not term[0] == ":":
            prefix, arg = term.split(":", 1)
            for meth, prefixes in self._prefixes.iteritems():
                if prefix in prefixes:
                    return meth(arg)
        return Text(self._parse_literal(term))

    def parse(self, query):
        """
        Parse a search query.

        The result is normalized with a sorting function so that ``"foo OR bar"``
        and ``"bar OR foo"`` result in the same tree. This is important for caching
        purposes.

        :param query: The query be converted.
        :type query: str

        :return: A tree storing the data in the query.
        :rtype: :py:class:`~.query.tree.Tree`

        :raises: :py:class:`.QueryParseException`
        """
        print "  STRING:", query
        for i, term in enumerate(split(query), 1):
            ## TODO: remove enumerate when removing debug prints
            print " in [%02d]:" % i, term
            node = self._parse_term(term)
            print "out [%02d]:" % i, node
            tree = Tree(node)
            print "    TREE:", tree
            return tree
            ## TODO

        # language:"Python"
        # lang:
        # l:

        # author:"Ben Kurtovic"

        # modified:before
        # modified:after
        # created:before
        # created:after:"Jaunary 4, 2014"

        # func:"foobar"
        # func:re|gex:"foo?b|car"

        # "foo" -> Tree()
        # "foo bar" -> "foo bar" OR ("foo" or "bar")
        # "foo bar baz" -> ""foo bar baz" OR ("foo" OR "bar baz") OR ("foo" OR "bar baz") OR ('foo' OR 'bar' OR 'baz')"


parse_query = _QueryParser().parse
